import os
import sys

import click

from ..core.instance import DagsterInstance
from ..version import __version__
from .instance import create_instance_cli_group
from .pipeline import create_pipeline_cli_group
from .run import create_run_cli_group
from .schedule import create_schedule_cli_group


def create_dagster_cli():
    commands = {
        'pipeline': create_pipeline_cli_group(),
        'run': create_run_cli_group(),
        'instance': create_instance_cli_group(),
        'schedule': create_schedule_cli_group(),
    }

    @click.group(commands=commands)
    @click.version_option(version=__version__)
    def group():
        'Noop'

    # add the path for the cwd so imports in dynamically loaded code work correctly
    sys.path.append(os.getcwd())
    return group


def main():
    cli = create_dagster_cli()
    cli(obj={})  # pylint:disable=E1123



import os
import sys

from dagster import DagsterEventType, execute_pipeline, lambda_solid, pipeline
from dagster.core.instance import DagsterInstance


@lambda_solid
def spew():
    print(HELLO_WORLD)
    return


@pipeline
def spew_pipeline():
    spew()


HELLO_WORLD = 'Hello World'
SEPARATOR = os.linesep if (os.name == 'nt' and sys.version_info < (3,)) else '\n'


def test_stdout():
    instance = DagsterInstance.local_temp()
    result = execute_pipeline(spew_pipeline, instance=instance)
    assert result.success
    compute_steps = [
        event.step_key
        for event in result.step_event_list
        if event.event_type == DagsterEventType.STEP_START
    ]
    assert len(compute_steps) == 1
    step_key = compute_steps[0]
    logs = instance.compute_log_manager.read_logs(result.run_id, step_key)
    assert logs.stdout.data == HELLO_WORLD + SEPARATOR

    cleaned_logs = logs.stderr.data.replace('\x1b[34m', '').replace('\x1b[0m', '')

    assert 'dagster - DEBUG - spew_pipeline - ' in cleaned_logs
    logs = instance.compute_log_manager.read_logs(result.run_id, step_key, cursor='0:0')
    assert logs.stdout.data == HELLO_WORLD + SEPARATOR
    assert 'dagster - DEBUG - spew_pipeline - ' in cleaned_logs
    assert instance.compute_log_manager.is_compute_completed(result.run_id, step_key)

    bad_logs = instance.compute_log_manager.read_logs('not_a_run_id', step_key)
    assert bad_logs.stdout.data is None
    assert bad_logs.stderr.data is None
    assert not instance.compute_log_manager.is_compute_completed('not_a_run_id', step_key)


def test_stdout_subscriptions():
    instance = DagsterInstance.local_temp()
    step_key = 'spew.compute'
    result = execute_pipeline(spew_pipeline, instance=instance)
    observable = instance.compute_log_manager.observable(result.run_id, step_key)
    logs = []
    observable.subscribe(logs.append)
    assert len(logs) == 1
    stdout_cursor, stderr_cursor = map(int, logs[0].cursor.split(':'))
    assert stdout_cursor == len(logs[0].stdout.data)
    assert stdout_cursor in [12, 13]
    assert stderr_cursor == len(logs[0].stderr.data)
    assert stderr_cursor > 400



import os
from collections import namedtuple
from enum import Enum

from dagster import check
from dagster.core.serdes import whitelist_for_serdes

from .utils import DEFAULT_OUTPUT


def last_file_comp(path):
    return os.path.basename(os.path.normpath(path))


@whitelist_for_serdes
class EventMetadataEntry(namedtuple('_EventMetadataEntry', 'label description entry_data')):
    '''A structure for describing metadata for Dagster events.

    Args:
        label (str):
        description (Optional[str]):
        entry_data (List[EntryDataUnion]):
            A list of typed metadata entries. The different types allow for customized display in
            tools like dagit.

    '''

    def __new__(cls, label, description, entry_data):
        return super(EventMetadataEntry, cls).__new__(
            cls,
            check.str_param(label, 'label'),
            check.opt_str_param(description, 'description'),
            check.inst_param(entry_data, 'entry_data', EntryDataUnion),
        )

    @staticmethod
    def text(text, label, description=None):
        'A EventMetadataEntry with a single TextMetadataEntryData entry'
        return EventMetadataEntry(label, description, TextMetadataEntryData(text))

    @staticmethod
    def url(url, label, description=None):
        'A EventMetadataEntry with a single UrlMetadataEntryData entry'
        return EventMetadataEntry(label, description, UrlMetadataEntryData(url))

    @staticmethod
    def path(path, label, description=None):
        'A EventMetadataEntry with a single PathMetadataEntryData entry'
        return EventMetadataEntry(label, description, PathMetadataEntryData(path))

    @staticmethod
    def fspath(path, label=None, description=None):
        'Just like path, but makes label last path component if None'
        return EventMetadataEntry.path(
            path, label if label is not None else last_file_comp(path), description
        )

    @staticmethod
    def json(data, label, description=None):
        return EventMetadataEntry(label, description, JsonMetadataEntryData(data))

    @staticmethod
    def md(mdString, label, description=None):
        return EventMetadataEntry(label, description, MarkdownMetadataEntryData(mdString))


@whitelist_for_serdes
class TextMetadataEntryData(namedtuple('_TextMetadataEntryData', 'text')):
    def __new__(cls, text):
        return super(TextMetadataEntryData, cls).__new__(cls, check.str_param(text, 'text'))


@whitelist_for_serdes
class UrlMetadataEntryData(namedtuple('_UrlMetadataEntryData', 'url')):
    def __new__(cls, url):
        return super(UrlMetadataEntryData, cls).__new__(cls, check.str_param(url, 'url'))


@whitelist_for_serdes
class PathMetadataEntryData(namedtuple('_PathMetadataEntryData', 'path')):
    def __new__(cls, path):
        return super(PathMetadataEntryData, cls).__new__(cls, check.str_param(path, 'path'))


@whitelist_for_serdes
class JsonMetadataEntryData(namedtuple('_JsonMetadataEntryData', 'data')):
    def __new__(cls, data):
        return super(JsonMetadataEntryData, cls).__new__(
            cls, check.dict_param(data, 'data', key_type=str)
        )


@whitelist_for_serdes
class MarkdownMetadataEntryData(namedtuple('_MarkdownMetadataEntryData', 'mdString')):
    def __new__(cls, mdString):
        return super(MarkdownMetadataEntryData, cls).__new__(
            cls, check.str_param(mdString, 'mdString')
        )


EntryDataUnion = (
    TextMetadataEntryData,
    UrlMetadataEntryData,
    PathMetadataEntryData,
    JsonMetadataEntryData,
    MarkdownMetadataEntryData,
)


class Output(namedtuple('_Result', 'value output_name')):
    '''A value produced by a solid compute function for downstream consumption. An implementer
    of a SolidDefinition directly must provide a compute function that yields objects of this type.

    Args:
        value (Any): Value returned by the compute function.
        output_name (str): Name of the output returns. Defaults to "result"
'''

    def __new__(cls, value, output_name=DEFAULT_OUTPUT):
        return super(Output, cls).__new__(cls, value, check.str_param(output_name, 'output_name'))


@whitelist_for_serdes
class Materialization(namedtuple('_Materialization', 'label description metadata_entries')):
    '''A value materialized by a solid compute function.

    As opposed to Outputs, Materializations can not be passed to other solids and persistence
    is not controlled by dagster. They are a useful way to communicate side effects to the system
    and display them to the end user.

    Args:
        label (str): A short display name for the materialized value.
        description (Optional[str]): A longer description of the materialized value.
        metadata_entries (Optional[List[EventMetadataEntry]]):
            Arbitrary metadata about the materialized value.
    '''

    @staticmethod
    def file(path, description=None):
        return Materialization(
            label=last_file_comp(path),
            description=description,
            metadata_entries=[EventMetadataEntry.fspath(path)],
        )

    def __new__(cls, label, description=None, metadata_entries=None):
        return super(Materialization, cls).__new__(
            cls,
            label=check.str_param(label, 'label'),
            description=check.opt_str_param(description, 'description'),
            metadata_entries=check.opt_list_param(
                metadata_entries, metadata_entries, of_type=EventMetadataEntry
            ),
        )


@whitelist_for_serdes
class ExpectationResult(
    namedtuple('_ExpectationResult', 'success label description metadata_entries')
):
    '''The result of a data quality test.

    ExpectationResults can be yielded from solids just like Outputs and Materializations.

    Args:
        success (bool): Whether the expectation passed or not.
        label (Optional[str]): Short display name for expectation. Defaults to "result".
        description (Optional[str]): A longer description of the data quality test.
        metadata_entries (Optional[List[EventMetadataEntry]]):
            Arbitrary metadata about the expectation.
    '''

    def __new__(cls, success, label=None, description=None, metadata_entries=None):
        return super(ExpectationResult, cls).__new__(
            cls,
            success=check.bool_param(success, 'success'),
            label=check.opt_str_param(label, 'label', 'result'),
            description=check.opt_str_param(description, 'description'),
            metadata_entries=check.opt_list_param(
                metadata_entries, metadata_entries, of_type=EventMetadataEntry
            ),
        )


@whitelist_for_serdes
class TypeCheck(namedtuple('_TypeCheck', 'description metadata_entries')):
    '''Used to communicate metadata about a value as it is evaluated against
    a declared expected type.
    '''

    def __new__(cls, description=None, metadata_entries=None):
        return super(TypeCheck, cls).__new__(
            cls,
            description=check.opt_str_param(description, 'description'),
            metadata_entries=check.opt_list_param(
                metadata_entries, metadata_entries, of_type=EventMetadataEntry
            ),
        )


class Failure(Exception):
    '''Can be raised from a solid compute function to return structured metadata
    about the failure.
    '''

    def __init__(self, description=None, metadata_entries=None):
        super(Failure, self).__init__(description)
        self.description = check.opt_str_param(description, 'description')
        self.metadata_entries = check.opt_list_param(
            metadata_entries, 'metadata_entries', of_type=EventMetadataEntry
        )


class ObjectStoreOperationType(Enum):
    SET_OBJECT = 'SET_OBJECT'
    GET_OBJECT = 'GET_OBJECT'
    RM_OBJECT = 'RM_OBJECT'
    CP_OBJECT = 'CP_OBJECT'


class ObjectStoreOperation(
    namedtuple(
        '_ObjectStoreOperation',
        'op key dest_key obj serialization_strategy_name object_store_name value_name',
    )
):
    '''Used internally by Dagster machinery when values are written to and read from an
    ObjectStore.

    Args:
        op (ObjectStoreOperationType): The type of the operation on the object store.
        key (str): The key of the object on which the operation was performed.
        dest_key (Optional[str]): The destination key, if any, to which the object was copied.
        obj (Any): The object, if any, retrieved by the operation.
        serialization_strategy_name (Optional[str]): The name of the serialization strategy, if any,
            employed by the operation
        object_store_name (Optional[str]): The name of the object store that performed the
            operation.
    '''

    def __new__(
        cls,
        op,
        key,
        dest_key=None,
        obj=None,
        serialization_strategy_name=None,
        object_store_name=None,
        value_name=None,
    ):
        return super(ObjectStoreOperation, cls).__new__(
            cls,
            op=op,
            key=check.str_param(key, 'key'),
            dest_key=check.opt_str_param(dest_key, 'dest_key'),
            obj=obj,
            serialization_strategy_name=check.opt_str_param(
                serialization_strategy_name, 'serialization_strategy_name'
            ),
            object_store_name=check.opt_str_param(object_store_name, 'object_store_name'),
            value_name=check.opt_str_param(value_name, 'value_name'),
        )

    @classmethod
    def serializable(cls, inst, **kwargs):
        return cls(
            **dict(
                {
                    'op': inst.op.value,
                    'key': inst.key,
                    'dest_key': inst.dest_key,
                    'obj': None,
                    'serialization_strategy_name': inst.serialization_strategy_name,
                    'object_store_name': inst.object_store_name,
                },
                **kwargs
            )
        )



from collections import namedtuple

from dagster import check
from dagster.core.definitions import ExecutorDefinition, ModeDefinition, PipelineDefinition
from dagster.core.execution.config import RunConfig
from dagster.core.system_config.objects import EnvironmentConfig


class InitExecutorContext(
    namedtuple(
        'InitExecutorContext',
        'pipeline_def mode_def executor_def run_config environment_config executor_config',
    )
):
    def __new__(
        cls, pipeline_def, mode_def, executor_def, run_config, environment_config, executor_config
    ):
        return super(InitExecutorContext, cls).__new__(
            cls,
            pipeline_def=check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition),
            mode_def=check.inst_param(mode_def, 'mode_def', ModeDefinition),
            executor_def=check.inst_param(executor_def, 'executor_def', ExecutorDefinition),
            run_config=check.inst_param(run_config, 'run_config', RunConfig),
            environment_config=check.inst_param(
                environment_config, 'environment_config', EnvironmentConfig
            ),
            executor_config=check.dict_param(executor_config, executor_config, key_type=str),
        )



import io
import os
import signal
import subprocess
import sys
import time
from contextlib import contextmanager

from dagster import check
from dagster.core.execution.context.system import SystemStepExecutionContext
from dagster.core.storage.compute_log_manager import ComputeIOType
from dagster.utils import ensure_file


@contextmanager
def mirror_step_io(step_context):
    check.inst_param(step_context, 'step_context', SystemStepExecutionContext)
    manager = step_context.instance.compute_log_manager
    if not manager.enabled(step_context):
        yield
        return

    outpath = manager.get_local_path(
        step_context.run_id, step_context.step.key, ComputeIOType.STDOUT
    )
    errpath = manager.get_local_path(
        step_context.run_id, step_context.step.key, ComputeIOType.STDERR
    )

    manager.on_compute_start(step_context)
    with mirror_stream(sys.stderr, errpath):
        with mirror_stream(sys.stdout, outpath):
            # compute function executed here
            yield
    manager.on_compute_finish(step_context)


@contextmanager
def mirror_stream(stream, path, buffering=1):
    ensure_file(path)
    with tailf(path):
        with open(path, 'a+', buffering=buffering) as to_stream:
            with redirect_stream(to_stream=to_stream, from_stream=stream):
                yield


@contextmanager
def redirect_stream(to_stream=os.devnull, from_stream=sys.stdout):
    # swap the file descriptors to capture system-level output in the process
    # From https://stackoverflow.com/questions/4675728/redirect-stdout-to-a-file-in-python/22434262#22434262
    from_fd = _fileno(from_stream)
    to_fd = _fileno(to_stream)

    if not from_fd or not to_fd:
        yield
        return

    with os.fdopen(os.dup(from_fd), 'wb') as copied:
        from_stream.flush()
        try:
            os.dup2(_fileno(to_stream), from_fd)
        except ValueError:
            with open(to_stream, 'wb') as to_file:
                os.dup2(to_file.fileno(), from_fd)
        try:
            yield from_stream
        finally:
            from_stream.flush()
            to_stream.flush()
            os.dup2(copied.fileno(), from_fd)


@contextmanager
def tailf(path):
    if sys.platform == 'win32':
        # no tail, bail
        yield
    else:
        # unix only
        cmd = 'tail -F -n 1 {}'.format(path).split(' ')

        # open a subprocess to tail the file and print to stdout
        tail_process = subprocess.Popen(cmd)

        # fork a child watcher process to sleep/wait for the parent process (which yields to the
        # compute function) to either A) complete and clean-up OR B) segfault / die silently.  In
        # the case of B, the spawned tail process will not automatically get terminated, so we need
        # to make sure that we terminate it explicitly and then exit.
        watcher_pid = os.fork()

        if watcher_pid == 0:
            # this is the child watcher process, sleep until orphaned, then kill the tail process
            # and exit
            while True:
                if os.getppid() == 1:  # orphaned process
                    time.sleep(1)
                    tail_process.terminate()
                    sys.exit()
                else:
                    time.sleep(1)
        else:
            # this is the parent process, yield to the compute function and then terminate both the
            # tail and watcher processes.
            try:
                yield
            finally:
                tail_process.terminate()
                os.kill(watcher_pid, signal.SIGTERM)


def _fileno(stream):
    try:
        fd = getattr(stream, 'fileno', lambda: stream)()
    except io.UnsupportedOperation:
        # Test CLI runners will stub out stdout to a non-file stream, which will raise an
        # UnsupportedOperation if `fileno` is accessed.  We need to make sure we do not error out,
        # or tests will fail
        return None

    if isinstance(fd, int):
        return fd

    return None
